from collections.abc import (
    AsyncGenerator,
    AsyncIterable,
    AsyncIterator,
    Iterable,
    Sequence,
)
from typing import TypeVar

T = TypeVar("T")


async def aenumerate(
    iterable: AsyncIterable[T], start: int = 0
) -> AsyncIterator[tuple[int, T]]:
    """Return an async iterator that yields tuples of (index, value)."""
    index = start
    async for item in iterable:
        yield index, item
        index += 1


async def amoving_window(
    iterable: AsyncIterable[T], max_length: int
) -> AsyncIterator[Sequence[T]]:
    """Return an async iterator moving a window of size max_length over iterable."""
    window = []

    async for item in iterable:
        window.append(item)
        if len(window) >= max_length:
            yield window
            window = []
    if window:
        yield window


async def as_async_generator(iterable: Iterable[T]) -> AsyncGenerator[T, None]:
    """Return an async generator from an iterable."""
    for item in iterable:
        yield item



import re
import shlex
from collections.abc import Callable
from typing import TypeVar

from typing_extensions import deprecated

__all__ = [
    "to_snake",
    "to_camel",
    "to_pascal",
    "to_kebab",
    "make_lex_separator",
    "comma_separator",
]

T = TypeVar("T")
OuterCastT = TypeVar("OuterCastT", list, tuple, set)

_to_camel_regexp = re.compile("(_|-)([a-zA-Z])")
_to_snake_regexp = re.compile("([a-z])([A-Z])")


def replace_all(string: str, replacements: dict[str, str]) -> str:
    for old, new in replacements.items():
        string = string.replace(old, new)
    return string


def to_snake(string: str) -> str:
    return replace_all(
        _to_snake_regexp.sub(r"\1_\2", string), {"-": "_", " ": "_"}
    ).lower()


def to_camel(string: str) -> str:
    return _to_camel_regexp.sub(
        lambda match: match[2].upper(), to_snake(string)
    ).rstrip("_-")


def to_pascal(string: str) -> str:
    val = to_camel(string)
    return val[0].upper() + val[1:]


upper_camel = deprecated('Use "to_pascal" instead.')(to_pascal)


def to_kebab(string: str, remove_trailing_underscores: bool = False) -> str:
    result = to_snake(string).replace("_", "-")
    return result if not remove_trailing_underscores else result.rstrip("-")


def make_lex_separator(
    outer_cast: type[OuterCastT], cast: type = str
) -> Callable[[str], OuterCastT]:
    def wrapper(value: str) -> OuterCastT:
        lex = shlex.shlex(value, posix=True)
        lex.whitespace = ","
        lex.whitespace_split = True
        return outer_cast(cast(item.strip()) for item in lex)

    return wrapper


def quote(string: str, quote_char: str = '"') -> str:
    return f"{quote_char}{string}{quote_char}"


comma_separator = make_lex_separator(tuple, str)



from typing import TypeVar

ExceptionT = TypeVar("ExceptionT", bound=Exception)


def sentence(exc_type: type[ExceptionT], message: str, *args) -> ExceptionT:
    """Add a period to the end of the message."""
    return exc_type(message.rstrip(".!?") + ".", *args)


def scream(exc_type: type[ExceptionT], message: str, *args) -> ExceptionT:
    """Add an exclamation point to the end of the message."""
    return exc_type(message.rstrip(".!?") + "!", *args)



__version__ = "0.1.0"

from . import exc, json, strings, timezone
from .asynciter import aenumerate, amoving_window, as_async_generator
from .casting import as_async, filter_isinstance, filter_issubclass, safe_cast
from .functions import cache, lazymethod
from .sequences import (
    flatten,
    indexsecond_enumerate,
    merge_dicts,
    moving_window,
    predicate_from_first,
)
from .worker import WorkerQueue

__all__ = [
    "json",
    "exc",
    "timezone",
    "aenumerate",
    "amoving_window",
    "as_async_generator",
    "filter_isinstance",
    "filter_issubclass",
    "as_async",
    "safe_cast",
    "lazymethod",
    "cache",
    "WorkerQueue",
    "moving_window",
    "flatten",
    "merge_dicts",
    "strings",
    "predicate_from_first",
    "indexsecond_enumerate",
]
